# pylint: disable=no-member
import sys

from google.cloud import bigquery
from google.oauth2 import service_account

from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.date_managers import date_range

class CustomGBQReader:
    """
    Custom Google BigQuery Console Reader
    """
    def __init__(self, service_account_secrets_path=None, config_path=None):

        if service_account_secrets_path is not None:
            self.credentials = load_file(service_account_secrets_path, fmt="json")

        if config_path is not None:
            self.config = load_file(config_path, fmt="yml")

    def get_auth(self):
        """
        Get our credentials initialised above and use those to get client

        """
        credentials = service_account.Credentials.from_service_account_info(self.credentials)
        client = bigquery.Client(
            credentials=credentials,
            project=credentials.project_id,
        )
        return client

    def build_query(self, query, date_run, metric):
        """ Build up query per according to metric """

        #initialization
        query=""
        project_id: str = self.credentials["project_id"]
        #Error Checking
        metrics_len = len(metric)

        query_prefix = "SELECT * FROM "
        table_name = f" {project_id}.analytics_150994723.events_{date_run} "
        parameters = "WHERE app_info.firebase_app_id = @firebase_id and event_name "
        if metrics_len == 0:
            print(f"please select atleast one metric to pull - current {metrics_len}")
            sys.exit(1)
        else:
            metrics = ''.join(metric)
            if metrics == 'session_start':
                query_suffix =  f" and event_name = '{metrics}'"
                query = query_prefix + table_name + parameters + query_suffix
            elif metrics == 'leads':
                query_prefix = "SELECT * FROM (SELECT *, (SELECT value.string_value"
                query_prefix += " FROM UNNEST(event_params) WHERE key='Category') AS "
                query_suffix = f"event_category FROM {table_name}) WHERE event_category='{metrics}'"
                query = query_prefix + query_suffix

        #TO DO - if more dimensions needed, add here
        return query

    def _query_handler(self, client, query, metrics=None, params=None):
        """ Query handler for the dataset
        """
        print(f"\n\n fetching data for: {metrics} \n\n")

        result=None
        if params is not None:
            job_config = bigquery.QueryJobConfig()
            job_config.query_parameters = params

            query_job = client.query(
                    query,
                    location='US',
                    job_config = job_config
                )
        # no parameters justs a simple query
        else:
            query_job = client.query(query)
        try:
            result = query_job.result()
            return result
        except SyntaxError as err:
            if err.code == 400:
                print(f"Syntax error: {err}")
                sys.exit(1)
            else:
                print(f"Got unexpected error:\n {err}")
                sys.exit(1)

    def run_query(self):
        """
        Consumes a config file and loops through the dims
        to return relevant data from Google Big Query.
        """
        _query=None
        client = self.get_auth()

        start_date: str = self.config["start_date"]
        end_date: str = self.config["end_date"]
        metric: str = self.config["metrics"]
        firebase_id: str = self.config["firebase_id"]
        dimensions: list = self.config["dimensions"]
        params = [bigquery.ScalarQueryParameter('firebase_id', 'STRING', firebase_id),]

        print(f"Gathering data between given dates {start_date} and {end_date}. ")
        # split request by date to reduce 504 errors
        #BigQuery tables are named in format (project_id)_name_YYYYMMDD
        for date in date_range(start_date=start_date, end_date=end_date):
            for dimension in dimensions :
                current_date=date.replace('-', '')
                query = self.build_query(_query, current_date, metric)

                print(f"Querying at date: {current_date}.")
                # run until none is returned or there is no more data in rows
                result = self._query_handler(
                    client,
                    query,
                    metric,
                    params
                )
                result = [dict(row.items()) for row in result]

                yield {"dimension":dimension, "metric" : metric, "date":date, "data":result}



import gzip
import json
from urllib.parse import urlparse

import boto3
from pyspark.sql import functions as f
from pyspark.sql import types as t


def convert(field):
    """
    Handles nested json strings that need to retain their structure.
    This also guarantees that data is not lost during schema inference.
    """
    if not isinstance(field.dataType, t.StructType) and not isinstance(
        field.dataType, t.ArrayType
    ):
        return field.name
    return f.to_json(field.name).alias(field.name)


def read_gz(file_url):
    """
    Reads gzipped metadata json file that has no .gz extension for spark to identify.
    """
    client = boto3.client("s3")

    url = urlparse(file_url)
    bucket, filepath = url.netloc, url.path.lstrip("/")

    obj = client.get_object(Bucket=bucket, Key=filepath)

    gzfile = gzip.decompress(obj["Body"].read())
    return json.loads(gzfile)


def filter_paths(bucket, prefix, position):
    """
    Gathers relevant paths in an s3 bucket and groups them by
    the specified position in the url.
    This can be used to read and write whilst retaining a specified
    level of file structuring of objects in buckets.

    Prefix can be used to speed up the search process.

    If the S3 urls are:
        s3://dl-base/metric-1/date-1
        s3://dl-base/metric-2/date-1
        s3://dl-base/metric-2/date-2
        s3://dl-base/metric-3/date-1

    One could set position to 4 and return:
        {
            'metric-1': ['s3://dl-base/metric-1/date-1'],
            'metric-2': ['s3://dl-base/metric-2/date-1', 's3://dl-base/metric-2/date-2'],
            'metric-3': ['s3://dl-base/metric-3/date-1']
        }

    """
    client = boto3.client("s3")
    paginator = client.get_paginator("list_objects")
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

    paths = []
    for page in page_iterator:
        keys = [key["Key"] for key in page["Contents"]]
        for key in keys:
            key_path = f"s3://{bucket}/{key}"
            if len(key_path.split(".")) > 1:
                paths.append(key_path)

    paths_dict = {}  # get all data paths associated with position
    for path in paths:
        identifier = path.split("/")[position]
        if identifier not in paths_dict:
            paths_dict[identifier] = []
        paths_dict[identifier].append(path)

    return paths_dict



# pylint: disable=too-few-public-methods
"""
    CUSTOM WRITER CLASSES
        - Class which manages writer tasks like
        auth, write metadata, write file, create dir structure
"""
import json
import os
from datetime import datetime

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, file_path, profile_name=None):
        self.bucket = bucket
        self.file_path = file_path

        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

        self.s3_resource = self.boto3_session.resource("s3")

    def write_partition_to_s3(self, json_data):
        """
        Partitions data by date to reduce duplication in s3.
        So the plan is to ensure that if we:
            - pull from a specific date, we overwrite all that days data in the buckets.
            - this way, we ensure all data is original and in its most updated state.
        """
        partition_dataset = {}
        for row in json_data:
            # each stat has its own relevant date field
            date_value = row.get("date", row.get("day", row.get("start_date")))
            if date_value not in partition_dataset:
                partition_dataset[date_value] = []
            partition_dataset[date_value].append(row)

        for date, date_dataset in partition_dataset.items():
            key_path = f"{self.file_path}/{date}.json"
            print(f"Writing data to {key_path}.")
            self.s3_resource.Object(self.bucket, key_path).put(
                Body=json.dumps(date_dataset)
            )

    def write_to_s3(self, json_data):
        """
        Writes json data to s3 and names it by
        date of writing. Also consumes expected metadata.json file.
        """
        timestamp_suffix = datetime.now().strftime("%Y_%m_%d_%H%M%S")
        key_path = f"{os.path.join(self.file_path, timestamp_suffix)}.json"

        self.s3_resource.Object(self.bucket, key_path).put(Body=json.dumps(json_data))



# pylint: disable=too-few-public-methods
# pylint: disable=no-member

"""
    CUSTOM WRITER CLASSES
"""
import json

import boto3


class CustomS3JsonWriter:
    """Class Extends Basic LocalGZJsonWriter"""

    def __init__(self, bucket, file_path, profile_name=None):
        self.bucket = bucket
        self.file_path = file_path
        self.profile_name = profile_name

        if profile_name is None:
            self.boto3_session = boto3.Session()
        else:
            self.boto3_session = boto3.Session(profile_name=profile_name)

    def write_partition_to_s3(self, json_data):
        """
        Partitions data by date to reduce duplication in s3.
        So the plan is to ensure that if we:
            - pull from a specific date, we overwrite all that days data in the buckets.
            - this way, we ensure all data is original and in its most updated state.
        """
        dimension = json_data["dimension"]
        date = json_data["date"].replace('-', '')
        dataset = json_data["data"]
        metric = json_data["metric"]
        key_path = f"{self.file_path}/{metric}/{dimension}/{date}.json"
        print(f"Writing data to {key_path}.")
        self.s3_resource.Object(self.bucket, key_path).put(
            Body=json.dumps(dataset)
        )



# pylint: disable=no-member,inconsistent-return-statements,wrong-import-order
import datetime
import os

# import ast
import json
import requests
from dateutil.relativedelta import relativedelta

from sdc_dp_helpers.api_utilities.retry_managers import request_handler
from sdc_dp_helpers.xero import auth_handler as pixie
from sdc_dp_helpers.xero.config_managers import get_config
from xero import Xero


def date_filter_helper(from_date: str, to_date: str, filter_field: str = None) -> str:
    """
    Custom implementation of date filters borrowed from:
    https://github.com/ClaimerApp/pyxero/blob/master/xero/basemanager.py
    """
    if not from_date:
        raise ValueError("No from_date set")

    # common date_field inside of the accounts and contacts modules is UpdatedDateUTC
    filter_field = "UpdatedDateUTC" if not filter_field else filter_field
    api_filter = filter_field + ">=DateTime(" + ",".join(from_date.split("-")) + ")"
    if to_date:
        api_filter = (
            api_filter
            + "&&"
            + filter_field
            + "<=DateTime("
            + ",".join(to_date.split("-"))
            + ")"
        )
    # end if

    return api_filter


class CustomXeroReader:
    """
    Custom Xero Reader
    """

    def __init__(self, **kwargs):
        self.config_path = kwargs.get("config_path")
        self.creds_path = kwargs.get("creds_path")
        self.config = get_config(self.config_path)
        three_months_back = (datetime.date.today() - relativedelta(months=3)).strftime(
            "%Y-%m-01"
        )
        today = datetime.date.today().strftime("%Y-%m-%d")

        self.config["from_date"] = self.config.get("start_date", three_months_back)
        self.config["to_date"] = self.config.get("end_date", today)
        self.from_date = None
        self.to_date = None

        self.client_id = self.config.get("client_id", None)
        if not self.client_id:
            raise ValueError("No client_id set")

        self.auth_token = pixie.get_auth_token(
            client_id=self.client_id, local_token_path=self.creds_path
        )

    def fetch_reports(self) -> dict:
        """
        Loops through reports in the config to pull each of them
        """
        data_set = {}
        for report_name in self.config.get("reports", []):
            if report_name not in [
                "BalanceSheet",
                "ProfitAndLoss",
                # "TrialBalance", #not pulling this atm
                "AgedPayablesByContact",
                "AgedReceivablesByContact",
            ]:
                raise ValueError(report_name + " is not supported or does not exist.")

            self.auth_token = pixie.get_auth_token(self.client_id, self.creds_path)

            self.auth_token.tenant_id = tenant_id = self.config.get("tenant_id")

            xero_obj = Xero(self.auth_token)
            trackingcategories = (
                i for i in xero_obj.trackingcategories.all() if i is not None
            )

            for tracking_category in trackingcategories:
                request_params = {
                    "report_name": report_name,
                    "tenant_id": tenant_id,
                    "tracking_category": tracking_category,
                    "xero_obj": xero_obj,
                    "auth_token": self.auth_token,
                    "from_date": self.config.get("from_date"),
                    "to_date": self.config.get("to_date"),
                }

                if report_name in ["BalanceSheet", "ProfitAndLoss"]:
                    result = self.fetch_report(request_params)
                    if result and result is not None:
                        data_set[
                            (
                                report_name
                                + "_"
                                + tracking_category["TrackingCategoryID"]
                            )
                        ] = result

                elif report_name in [
                    "AgedPayablesByContact",
                    "AgedReceivablesByContact",
                ]:

                    for contact in xero_obj.contacts.filter(
                        raw="AccountNumber!=null", AccountNumber__startswith="999999"
                    ):
                        print(contact["ContactID"])
                        request_params.update({"contactId": contact["ContactID"]})
                        # print(type(report_name),
                        # type(tracking_category["TrackingCategoryID"]), type(contact_id))

                        result = self.fetch_report(request_params)
                        if result and result is not None:
                            data_set[
                                (
                                    report_name
                                    + "_"
                                    + tracking_category["TrackingCategoryID"]
                                    + "_"
                                    + contact["ContactID"]
                                )
                            ] = result
                else:
                    raise ValueError(
                        f"Unknown report named '{report_name}'. Check your spelling maybe?"
                    )

        return data_set

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def fetch_report(self, request_params: dict) -> str:
        """
        This method accepts Parameters
            report_name: report name as per Xero API endpoint
            from_date: start_date of the report
            to_date: end date of the report
            returns: report for those parameters using the requests API directly
        """
        # unpack request
        report_name = request_params["report_name"]

        today = datetime.date.today()
        three_months_back = (today - relativedelta(months=3)).strftime("%Y-%m-01")
        today = today.strftime("%Y-%m-%d")

        self.from_date = request_params.get(
            "from_date", request_params.get("date", three_months_back)
        )
        self.to_date = request_params.get("to_date", today)
        tracking_category = request_params["tracking_category"]

        self.auth_token = pixie.get_auth_token(
            client_id=self.client_id, local_token_path=self.creds_path
        )
        # self.auth_token.tenant_id = request['tenant_id']

        my_headers = {
            "Authorization": "Bearer "
            + self.auth_token.token[
                "access_token"
            ],  # self.xero_client.token["access_token"],
            "Xero-Tenant-Id": request_params["tenant_id"],
            "Accept": "application/json",
        }
        my_params = {
            "fromDate": self.from_date,
            "toDate": self.from_date,
        }
        if tracking_category:
            my_params.update(
                {"trackingCategoryID": tracking_category["TrackingCategoryID"]}
            )

        if "filter_items" in request_params.keys():
            my_params.update(request_params["filter_items"])

        if "contactId" in request_params.keys():
            my_params.update({"contactId": request_params["contactId"]})

        response = requests.get(
            "https://api.xero.com/api.xro/2.0/Reports/" + report_name,
            params=my_params,
            headers=my_headers,
            timeout=30
        )
        # response.text = response.text.strip("'<>() ").replace('\'', '\"')
        # print(response.text)
        report = json.loads(
            response.text.replace("\r", "").replace("\n", "").strip("'<>() ")
        )
        report = {
            "tenantId": self.auth_token.tenant_id,
            "trackingCategoryId": tracking_category["TrackingCategoryID"],
            "report": report,
            "from_date": self.from_date,
            "to_date": self.from_date,
        }
        return json.dumps(report)

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0.1)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    def run_request(self, xero_client, api_object, request):
        """
        Run the API request that consumes a request payload and site url.
        This separates the request with the request handler from the rest of the logic.
        """
        # To Do Handle API Errors
        api_call = getattr(xero_client, api_object)
        # XeroRateLimitExceeded
        return api_call.filter(
            raw=date_filter_helper(request["from_date"], request["to_date"]),
            page=request["page"],
        )

    def fetch_modules(self) -> dict:
        """
        Consumes a .yaml config file and loops through the date and url
        to return relevant data from Xero API.
        """
        self.auth_token = pixie.get_auth_token(self.client_id, self.creds_path)
        self.auth_token.tenant_id = [
            i["tenantId"]
            for i in self.auth_token.get_tenants()
            if i["tenantId"] == self.config["tenant_id"]
        ][0]
        xero = Xero(self.auth_token)

        today = datetime.date.today()
        three_months_back = (today - relativedelta(months=3)).strftime("%Y-%m-01")
        today = today.strftime("%Y-%m-%d")

        self.from_date = self.config.get("from_date", three_months_back)
        self.to_date = self.config.get("to_date", today)
        data_set = {}

        for api_object in self.config.get("modules", []):
            if api_object not in [
                "contacts",
                "accounts",
                "invoices",
                "banktransactions",
                "manualjournals",
                "purchaseorders",
            ]:
                raise ValueError(api_object + " is not supported or does not exist.")
            data_set[api_object] = []
            prev_response = None
            page = 1

            while True:
                response = self.run_request(
                    xero_client=xero,
                    api_object=api_object,
                    request={
                        "from_date": self.from_date,
                        "to_date": self.to_date,
                        "page": page,
                    },
                )
                if len(response) < 1:
                    print("Request returned empty payload. breaking...")
                    break
                if response == prev_response:
                    print("Request returned copy of last payload. breaking...")
                    break
                data_set[api_object] += [
                    # the response objects are returned as a dictionary with datetime.
                    # datetime objects,to convert to json type we ask json to attempt to
                    # convert all pythonic objects to something acceptable for a json object
                    # only thereafter can we convert back to json for saving to json file
                    json.loads(
                        json.dumps(response_obj, indent=4, sort_keys=True, default=str)
                    )
                    for response_obj in response
                ]

                # ensure the token is still fresh
                self.auth_token = pixie.get_auth_token(self.client_id, self.creds_path)
                prev_response = response
                page += 1
        return data_set

    def run_query(self):
        """
        As dictated by the config;
        This function fetches all modules and reports requested based on the config
        """
        if getattr(self.config, "modules", None):
            return self.fetch_modules()
        if getattr(self.config, "reports", None):
            return self.fetch_reports()



from datetime import timedelta, datetime

from sdc_dp_helpers.api_utilities.date_managers import date_handler


def date_range(start_date, end_date, delta=timedelta(days=1)):
    """
    The range is inclusive, so both start_date and end_date will be returned.
    :start_date: The datetime object representing the first day in the range.
    :end_date: The datetime object representing the second day in the range.
    :delta: A datetime.timedelta instance, specifying the step interval. Defaults to one day.
    Yields:
        Each datetime object in the range.
    """

    start_date = date_handler(date_string=start_date)
    end_date = date_handler(date_string=end_date)

    start_date = datetime.strptime(start_date, "%Y-%m-%d")
    end_date = datetime.strptime(end_date, "%Y-%m-%d")

    current_date = start_date
    while current_date <= end_date:
        yield current_date.strftime("%Y-%m-%d")
        current_date += delta


def un_nest_keys(data, col, key_list, value_list):
    """
    A simple method that takes a set of keys and values
    from a dataset with dimensions and replaces them
    with key value pairs.

    :data: The dataset that needs to be un_nested.
    :col: The key name of the dimensions.
    :key_list: List of keys.
    :value_list: List of Values.

    returns:
        So dataset A:
            {'keys': [1, 2, 3]}
        Can become:
            {'key1': 1, 'key2': 2, 'key3': 3}
    """
    data.update(dict(zip(key_list, value_list)))
    del data[col]
    return data



# pylint: disable=no-member,too-many-locals
"""
    CUSTOM READERS CLASSES
        - Class which manages reader tasks like auth, requests, pagination
"""
import datetime
import os
import socket
import sys
from typing import Callable, List, Dict

import httplib2
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from oauth2client.service_account import ServiceAccountCredentials

from sdc_dp_helpers.api_utilities.date_managers import date_range
from sdc_dp_helpers.api_utilities.file_managers import load_file
from sdc_dp_helpers.api_utilities.retry_managers import request_handler, retry_handler
from sdc_dp_helpers.api_utilities.tracking_metadata import tracking_metadata


class GAV3Reader:
    """
    Google analytics v3
    """
    def __init__(
        self, service_account_email, creds_filepath, config_filepath, **kwargs
    ):
        self.service_account_email = service_account_email
        self.creds_filepath = creds_filepath
        self.scopes = kwargs.get(
            "scopes", ["https://www.googleapis.com/auth/analytics.readonly"]
        )

        self.config = load_file(config_filepath, "yaml")
        self.tracking_object = []

    def _get_service(self):
        """
        Get a service that communicates to a Google API.
        """
        credentials = ServiceAccountCredentials.from_p12_keyfile(
            filename=self.creds_filepath,
            scopes=self.scopes,
            service_account_email=self.service_account_email,
        )
        service = build(serviceName="analytics", version="v3", credentials=credentials)
        return service

    @staticmethod
    def _normalize_data(report):
        """
        Flattens the response data to a standard dict.
        """
        headers = [head["name"] for head in report["columnHeaders"]]
        profile = report["profileInfo"]
        data_set = []
        for row in report["rows"]:
            tmp = dict(zip(headers, row))
            tmp.update(profile)
            data_set.append(tmp)
        return data_set

    @request_handler(wait=1, backoff_factor=0.5)
    @retry_handler(
        exceptions=(socket.timeout, HttpError),
        total_tries=5,
        initial_wait=60,
        backoff_factor=2,
    )
    def _query_handler(self, view_id, start_index, date):
        """
        Separated query method to handle retry and delay methods.
        """
        print(f"Querying View ID: {view_id}. At index: {start_index}.")
        # frequently rebuild the service
        service = self._get_service()
        response = (
            service.data()
            .ga()
            .get(
                ids=f"ga:{view_id}",
                start_date=date,
                end_date=date,
                # convert mets and dims into comma separated string
                metrics=",".join(self.config.get("metrics", [])),
                dimensions=",".join(self.config.get("dimensions", [])),
                sort=self.config.get("sort", None),
                filters=self.config.get("filters", None),
                max_results=self.config.get("max_results", 10000),
                start_index=start_index,
                samplingLevel="HIGHER_PRECISION",
                include_empty_rows=True,
            )
        )

        # tracking reader progress
        self.tracking_object = tracking_metadata(
            data_object=self.tracking_object,
            service="google_analytics_v3",
            response_datetime=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            index=start_index,
            id_name="view_id",
            id_value=view_id,
            config_datetime=self.config.get("start_date"),
            configuration=self.config,
            byte_size=sys.getsizeof(response),
        )

        return response.execute()

    def run_query(self):
        """
        Core V3 Reporting API.
        The method will try each query 5 times before failing.
        Config File:
            start_date: |
              Start date for fetching Analytics data.
              Requests can specify a start date formatted as YYYY-MM-DD,
              or as a relative date (e.g., today, yesterday, or
              NdaysAgo where N is a positive integer).
            end_date: |
              End date for fetching Analytics data.
              Request can specify an end date formatted as YYYY-MM-DD,
              or as a relative date (e.g., today, yesterday, or
              NdaysAgo where N is a positive integer).
            metrics: A maximum of 10 Metrics in list form.
            dimensions: A maximum of 7 Dimensions in list form.
            sort: |
              A list of comma-separated dimensions and metrics
              indicating the sorting order and sorting direction for the returned data.
            filters: Dimension or metric filters that restrict the data returned for your request.
            view_ids: |
              A list of the unique table ID of the form ga:XXXX, where XXXX is the
              Analytics view (profile) ID for which the query will retrieve the data.
        """
        for view_id in self.config.get("view_ids"):
            for date in date_range(
                self.config.get("start_date"), self.config.get("end_date")
            ):
                date_dataset = []
                print(f"Gathering data for {view_id} at date: {date}.")
                page_token = 1  # handle paging of response
                while True:
                    response = self._query_handler(
                        view_id=view_id,
                        start_index=page_token,
                        date=date,
                    )
                    if not response:
                        print(f"No response for view_id: {view_id}")
                        break
                    if len(response.get("rows", [])) < 1:
                        print("No more data in response.")
                        break
                    page_token += response.get("itemsPerPage", 0)

                    date_dataset.extend(self._normalize_data(response))

                # generator preserves memory, additional data provided
                # to efficiently partition to s3
                # only yield if data is returned.
                if date_dataset:
                    # TO DO: use self.tracking_object byte_size attrib
                    #       to determine whether or not to yield here
                    yield {"data": date_dataset, "date": date, "view_id": view_id}
                else:
                    print("View ID had no data for given date, skipping.")


class BaseGAV4Reader:
    """
    Base Google Analytics V4 reader class.

    Used as basis for building custom GA V4 readers, it requires that you
    add your own Authentication through the auth_credentials method.
    The basic workflow is:
    (1) Create credentials
    (2) Build google service object
    (3) Return an iterable of the get_report method
    (4) Iterate until no pagination token is found. A.k.a. pageToken==None
    """

    analytics = None
    max_pages = None

    @staticmethod
    def _normalize_gav4(response: List[Dict], view_id: str) -> list:
        for idx, report in enumerate(response):
            if idx < len(report):
                column_header = report.get("columnHeader", {})
                dimension_headers = column_header.get("dimensions", [])
                metric_headers = column_header.get("metricHeader", {}).get(
                    "metricHeaderEntries", []
                )
                rows = report.get("data", {}).get("rows", [])
                dataset = []
                for row in rows:
                    row_data = {}
                    dimensions = row.get("dimensions", [])
                    date_range_values = row.get("metrics", [])

                    for header, dimension in zip(dimension_headers, dimensions):
                        row_data[header] = dimension
                    for  values in enumerate(date_range_values):
                        for metric_header, value in zip(
                            metric_headers, values.get("values")
                        ):
                            row_data[metric_header.get("name")] = value
                        row_data["view_id"] = view_id
                    dataset.append(row_data)
        return dataset

    class ReaderIter:
        """Inner Class - Iterator that yields queries until max pages or pageToken == None."""

        page_counter = 0

        def __init__(
            self, *, query_fn: Callable, query: list, page_token: str = None, **kwargs
        ):
            self.query_fn = query_fn
            self.max_pages = kwargs.get("max_pages", None)
            self.page_token = page_token
            self.query = query

        def __iter__(self):
            """Iterator method for iterator class"""
            return self

        def __next__(self):
            """
            Next method for iterator class which includes
            the StopIteration criteria for iterable
            """
            # check StopIteration criteria
            if self.max_pages is not None:
                if self.page_counter >= self.max_pages:
                    print(f"stopping iteration on max pages = {self.max_pages}")
                    raise StopIteration

            if self.page_counter > 0 and self.page_token is None:
                print(f"stopping iteration on pageToken = {self.page_token}")
                raise StopIteration

            # run get_report
            report, self.page_token = self.query_fn(
                queries=self.query, page_token=self.page_token
            )
            page=self.page_counter + 1
            view_id=self.query.get("viewId")
            print(
                f"Finished query function for page {page} of view ID {view_id}"
            )

            # increment pagination counter
            self.page_counter += 1

            return report

    # pylint: disable=no-member
    def build_analytics_service_object(
        self, api_name: str = "analyticsreporting", api_version: str = "v4"
    ):
        """Initializes the analytics reporting service object.
        args:
        client_secrets_path (json) : Google API OAuth2 API credentials
        scopes (list or iterable) : scopes for Google OAuth2 authentication

        Returns:
        analytics an authorized analyticsreporting service object.
        """
        # authorize HTTP object
        http = self.credentials.authorize(http=httplib2.Http())
        # Build the service object.
        return build(api_name, api_version, http=http)

    @request_handler(
        wait=int(os.environ.get("REQUEST_WAIT_TIME", 0)),
        backoff_factor=float(os.environ.get("REQUEST_BACKOFF_FACTOR", 0.01)),
        backoff_method=os.environ.get("REQUEST_BACKOFF_METHOD", 0.01),
    )
    @retry_handler(
        exceptions=(socket.timeout, HttpError),
        total_tries=5,
        initial_wait=60,
        backoff_factor=2,
    )
    def get_report(self, queries: list, page_token: str = None):
        """
        Use the Analytics Service Object to query the Analytics Reporting API V4.

        args:
        analytics : analytics reporting service object
        query (list): array of queries. [max. 5]
        page_token (str): token sent denoting the last scraped data point
        return:
        analytics report data (json)

        """
        if page_token is not None:
            for query in queries:
                query.update({"page_token": page_token})

        # run batch query
        report = (
            self.analytics.reports()
            .batchGet(body={"reportRequests": queries})
            .execute()
            .get("reports")
        )
        page_token = report[0].get("nextPageToken")

        return report, page_token

    def _query_handler(self, query: dict, page_token: str = None):
        date_dataset = []
        view_id = query.get("view_id")

        # """
        # returns a sync iterator
        # """
        iter_ = self.ReaderIter(
            query_fn=self.get_report,
            query=query,
            max_pages=self.max_pages,
            page_token=page_token,
        )
        for response in enumerate(iter_):
            if not response:
                print(f"No response for view_id: {view_id}")
                break
            for values in enumerate(response):
                date_dataset.extend(self._normalize_gav4(response, view_id))
                print(f"length of dataset is {len(date_dataset)}")
                token = values.get("nextPageToken")
                if not token:
                    break
                page_token = token
                print(f"page token is {page_token}")

        return date_dataset

    def run_query(self):
        date_dataset: list = []
        for view_id in self.config.get("view_ids"):
            for date in date_range(
                start_date=self.config.get("start_date"),
                end_date=self.config.get("end_date"),
            ):
                print(f"Gathering data for {view_id} at date: {date}.")
                query: dict = self.build_query(self.config, date, view_id)
                date_dataset: list[Dict] = self._query_handler(query, page_token="1")
                # generator preserves memory, additional data provided
                # to efficiently partition to s3
                # only yield if data is returned.
                if date_dataset:
                    # TO DO: use self.tracking_object byte_size attrib
                    #       to determine whether or not to yield here
                    yield {"data": date_dataset, "date": date, "view_id": view_id}
                else:
                    print("View ID had no data for given date, skipping.")


class CustomGoogleAnalyticsReaderWithServiceAcc(BaseGAV4Reader):
    """
    Custom Google Analytics Reader which authenticates with Service account
    """

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        service_account_secrets_path: str,
        service_account_email: str,
        scopes: list,
        **kwargs,
    ):
        self.max_pages = kwargs.get("max_pages", None)
        self.credentials = self.auth_credentials(
            service_account_secrets_path, service_account_email, scopes
        )
        # build analytics object
        self.analytics = self.build_analytics_service_object()
        self.config = load_file(kwargs.get("config_filepath"), "yaml")

    @staticmethod
    def auth_credentials(service_account_secrets_path, service_account_email, scopes):
        """Get a service that communicates to a Google API.
        Args:
            service_account_secrets_path: The filepath to service secrets.
            api_name: The name of the api to connect to.
            api_version: The api version to connect to.
            scope: A list auth scopes to authorize for the application.
            key_file_location: The path to a valid service account p12 key file.
            service_account_email: The service account email address.
            scopes: Allowed scopes for the given credentials
        Returns:
            A service that is connected to the specified API.
        """
        return ServiceAccountCredentials.from_p12_keyfile(
            service_account_email, service_account_secrets_path, scopes=scopes
        )

    @staticmethod
    def build_query(config: dict, date: str, view_id: str) -> List[Dict]:
        query = {
            "dateRanges": {"startDate": date, "endDate": date},
            "metrics": [{"expression": metric} for metric in config.get("metrics", [])],
            "dimensions": [
                {"name": dimension} for dimension in config.get("dimensions", [])
            ],
            "viewId": view_id,
        }

        return [query]
